from dataclasses import dataclass
from typing import Any
from typing import List
from typing import Tuple
from typing import Union

from pyaaware import FeatureGenerator

from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosF
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import Augmentation
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import ClassCount
from sonusai.mixture.types import DataClassSonusAIMixin
from sonusai.mixture.types import Feature
from sonusai.mixture.types import GeneralizedIDs
from sonusai.mixture.types import ImpulseResponseData
from sonusai.mixture.types import ImpulseResponseFiles
from sonusai.mixture.types import Location
from sonusai.mixture.types import NoiseFiles
from sonusai.mixture.types import Optional
from sonusai.mixture.types import Segsnr
from sonusai.mixture.types import TargetFiles
from sonusai.mixture.types import Truth
from sonusai.mixture.types import TruthSettings


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class MPGlobal:
    noise_augmentations: Augmentations = None
    ir_data: List[ImpulseResponseData] = None


MP_GLOBAL = MPGlobal()


@dataclass
class MRecord(DataClassSonusAIMixin):
    name: Location = None
    noise_augmentation_index: int = None
    noise_file_index: int = None
    noise_offset: int = None
    noise_snr_gain: float = None
    random_snr: Optional[bool] = None
    samples: int = None
    snr: float = None
    target_augmentation_index: List[int] = None
    target_file_index: List[int] = None
    target_gain: List[int] = None
    target_snr_gain: float = None


MRecords = List[MRecord]


@dataclass
class MixtureDatabaseConfig(DataClassSonusAIMixin):
    class_balancing: Optional[bool] = False
    class_balancing_augmentation: Optional[Augmentation] = None
    class_labels: List[str] = None
    class_weights_threshold: List[float] = None
    exhaustive_noise: Optional[bool] = True
    feature: str = None
    first_cba_index: Optional[int] = None
    ir_files: ImpulseResponseFiles = None
    mixtures: MRecords = None
    noise_augmentations: Augmentations = None
    noises: NoiseFiles = None
    num_classes: int = None
    random_snrs: Optional[List[str]] = None
    seed: Optional[int] = 0
    snrs: List[float] = None
    target_augmentations: Augmentations = None
    targets: TargetFiles = None
    truth_mutex: bool = None
    truth_reduction_function: str = None
    truth_settings: TruthSettings = None


@dataclass
class TransformConfig:
    N: int
    R: int
    ttype: str


class MixtureDatabase:
    from sonusai.mixture.types import AudiosT
    from sonusai.mixture.types import ListAudiosT

    def __init__(self,
                 config: Union[Location, MixtureDatabaseConfig],
                 lazy_load: bool = False,
                 show_progress: bool = False):
        if isinstance(config, MixtureDatabaseConfig):
            self.config = config
            self.location = None
        else:
            self._load_from_location(config)
            self.location = config

        self.fg = FeatureGenerator(feature_mode=self.feature,
                                   num_classes=self.num_classes,
                                   truth_mutex=self.truth_mutex)

        self.ft_config = TransformConfig(N=self.fg.ftransform_N,
                                         R=self.fg.ftransform_R,
                                         ttype=self.fg.ftransform_ttype)

        self.eft_config = TransformConfig(N=self.fg.eftransform_N,
                                          R=self.fg.eftransform_R,
                                          ttype=self.fg.eftransform_ttype)

        self.it_config = TransformConfig(N=self.fg.itransform_N,
                                         R=self.fg.itransform_R,
                                         ttype=self.fg.itransform_ttype)

        self.show_progress = show_progress
        self._ir_data = None
        self._raw_target_audios = None
        self._augmented_noise_audios = None

        if not lazy_load:
            self.load_ir_data(show_progress=show_progress)
            self.load_raw_target_audios(show_progress=show_progress)
            self.load_augmented_noise_audios(show_progress=show_progress)

    def _load_from_location(self, location: Location) -> None:
        import json
        from os.path import exists
        from os.path import isdir
        from os.path import join

        from sonusai import SonusAIError

        if not isdir(location):
            raise SonusAIError(f'{location} is not a directory')

        filename = join(location, 'mixdb.json')
        if not exists(filename):
            raise SonusAIError(f'could not find mixture database in {location}')

        with open(file=filename, mode='r', encoding='utf-8') as f:
            self.config = MixtureDatabaseConfig.from_dict(json.loads(f.read()))

    @property
    def json(self) -> str:
        """Convert MixtureDatabase to JSON

        :return: JSON representation of database
        """
        return self.config.to_json(indent=2)

    def save(self) -> None:
        """Save the MixtureDatabase as a JSON file

        """
        from os.path import join

        json_name = join(self.location, 'mixdb.json')
        with open(file=json_name, mode='w') as file:
            file.write(self.json)

    def forward_transform(self, audio: AudioT) -> AudioF:
        """Tranform time domain data into frequency domain using the forward transform config from the feature

        A new transform is used for each call; i.e., state is not maintained between calls to forward_transform().

        :param audio: Time domain data [samples]
        :return: Frequency domain data [frames, bins]
        """
        from pyaaware import ForwardTransform

        from sonusai.mixture import calculate_transform_from_audio

        return calculate_transform_from_audio(audio=audio,
                                              transform=ForwardTransform(N=self.ft_config.N,
                                                                         R=self.ft_config.R,
                                                                         ttype=self.ft_config.ttype))

    def inverse_transform(self, transform: AudioF, trim: bool = True) -> AudioT:
        """Tranform frequency domain data into time domain using the inverse transform config from the feature

        A new transform is used for each call; i.e., state is not maintained between calls to inverse_transform().

        :param transform: Frequency domain data [frames, bins]
        :param trim: Removes starting samples so output waveform will be time-aligned with input waveform to the transform
        :return: Time domain data [samples]
        """
        from pyaaware import InverseTransform

        from sonusai.mixture import calculate_audio_from_transform

        return calculate_audio_from_transform(data=transform,
                                              transform=InverseTransform(N=self.it_config.N,
                                                                         R=self.it_config.R,
                                                                         ttype=self.it_config.ttype),
                                              trim=trim)

    @property
    def class_balancing(self):
        return self.config.class_balancing

    @class_balancing.setter
    def class_balancing(self, value):
        self.config.class_balancing = value

    @property
    def class_balancing_augmentation(self):
        return self.config.class_balancing_augmentation

    @class_balancing_augmentation.setter
    def class_balancing_augmentation(self, value):
        self.config.class_balancing_augmentation = value

    @property
    def class_labels(self):
        return self.config.class_labels

    @class_labels.setter
    def class_labels(self, value):
        from sonusai import SonusAIError

        if value is not None and (not isinstance(value, list) or len(value) != self.num_classes):
            raise SonusAIError(f'invalid class_labels length')
        self.config.class_labels = value

    @property
    def class_weights_threshold(self):
        return self.config.class_weights_threshold

    @class_weights_threshold.setter
    def class_weights_threshold(self, value):
        from sonusai import SonusAIError

        if len(value) not in [1, self.num_classes]:
            raise SonusAIError(f'invalid class_weights_threshold length: {len(class_weights_threshold)}')
        self.config.class_weights_threshold = value

    @property
    def exhaustive_noise(self):
        return self.config.exhaustive_noise

    @exhaustive_noise.setter
    def exhaustive_noise(self, value):
        self.config.exhaustive_noise = value

    @property
    def feature(self):
        return self.config.feature

    @feature.setter
    def feature(self, value):
        self.config.feature = value

    @property
    def transform_frame_ms(self):
        from sonusai.mixture import SAMPLE_RATE

        return float(self.ft_config.R) / float(SAMPLE_RATE / 1000)

    @property
    def feature_ms(self):
        return self.transform_frame_ms * self.fg.decimation * self.fg.stride

    @property
    def feature_samples(self):
        return self.ft_config.R * self.fg.decimation * self.fg.stride

    @property
    def feature_step_ms(self):
        return self.transform_frame_ms * self.fg.decimation * self.fg.step

    @property
    def feature_step_samples(self):
        return self.ft_config.R * self.fg.decimation * self.fg.step

    @property
    def first_cba_index(self):
        return self.config.first_cba_index

    @first_cba_index.setter
    def first_cba_index(self, value):
        self.config.first_cba_index = value

    @property
    def ir_files(self):
        return self.config.ir_files

    @property
    def mixtures(self):
        return self.config.mixtures

    @mixtures.setter
    def mixtures(self, value):
        self.config.mixtures = value

    @property
    def noise_augmentations(self):
        return self.config.noise_augmentations

    @noise_augmentations.setter
    def noise_augmentations(self, value):
        self.config.noise_augmentations = value

    @property
    def noises(self):
        return self.config.noises

    @noises.setter
    def noises(self, value):
        self.config.noises = value

    @property
    def num_classes(self):
        return self.config.num_classes

    @num_classes.setter
    def num_classes(self, value):
        self.config.num_classes = value

    @property
    def random_snrs(self):
        return self.config.random_snrs

    @random_snrs.setter
    def random_snrs(self, value):
        self.config.random_snrs = value

    @property
    def seed(self):
        return self.config.seed

    @seed.setter
    def seed(self, value):
        self.config.seed = value

    @property
    def snrs(self):
        return self.config.snrs

    @snrs.setter
    def snrs(self, value):
        self.config.snrs = value

    @property
    def target_augmentations(self):
        return self.config.target_augmentations

    @target_augmentations.setter
    def target_augmentations(self, value):
        self.config.target_augmentations = value

    @property
    def targets(self):
        return self.config.targets

    @targets.setter
    def targets(self, value):
        self.config.targets = value

    @property
    def truth_mutex(self):
        return self.config.truth_mutex

    @truth_mutex.setter
    def truth_mutex(self, value):
        self.config.truth_mutex = value

    @property
    def truth_reduction_function(self):
        return self.config.truth_reduction_function

    @truth_reduction_function.setter
    def truth_reduction_function(self, value):
        self.config.truth_reduction_function = value

    @property
    def truth_settings(self):
        return self.config.truth_settings

    @truth_settings.setter
    def truth_settings(self, value):
        self.config.truth_settings = value

    @property
    def raw_target_audios(self) -> AudiosT:
        """Get the list of raw (unaugmented) target audio, loading from disk if necessary

        :return: List of raw target audio
        """
        if self._raw_target_audios is None:
            self.load_raw_target_audios(show_progress=self.show_progress)

        return self._raw_target_audios

    def load_raw_target_audios(self, show_progress: bool = False) -> None:
        """Load all raw (unaugmented) target audio into memory

        :param show_progress: Enable progress bar display
        """
        from tqdm import tqdm

        from sonusai.mixture import read_audio
        from sonusai.utils import p_tqdm_map

        names = [target.name for target in self.targets]
        progress = tqdm(total=len(names), desc='Read target audio', disable=not show_progress)
        self._raw_target_audios = p_tqdm_map(read_audio, names, progress=progress)
        progress.close()

    @property
    def augmented_noise_audios(self) -> ListAudiosT:
        """Get the list of augmented noise audio, loading from disk if necessary

        :return: List of augmented noise audio
        """
        if self._augmented_noise_audios is None:
            self.load_augmented_noise_audios(show_progress=self.show_progress)

        return self._augmented_noise_audios

    def load_augmented_noise_audios(self, show_progress: bool = False) -> None:
        """Load all augmented noise audio into memory

        :param show_progress: Enable progress bar display
        """
        from tqdm import tqdm

        from sonusai.mixture import read_audio
        from sonusai.utils import p_tqdm_map

        names = [noise.name for noise in self.noises]

        progress = tqdm(total=len(names), desc='Read noise audio', disable=not show_progress)
        raw_noise_audios = p_tqdm_map(read_audio, names, progress=progress)
        progress.close()

        MP_GLOBAL.noise_augmentations = self.noise_augmentations
        MP_GLOBAL.ir_data = self.ir_data

        progress = tqdm(total=len(names), desc='Augment noise audio', disable=not show_progress)
        self._augmented_noise_audios = p_tqdm_map(_augment_noise_audio, raw_noise_audios, progress=progress)
        progress.close()

    @property
    def ir_data(self) -> List[ImpulseResponseData]:
        """Get the list of impulse response data, loading from disk if necessary

        :return: List of impulse response data
        """
        if self._ir_data is None:
            self.load_ir_data(show_progress=self.show_progress)

        return self._ir_data

    def load_ir_data(self, show_progress: bool = False) -> None:
        """Load all impulse response data into memory

        :param show_progress: Enable progress bar display
        :return: List of impulse response data
        """
        if len(self.ir_files) == 0:
            self._ir_data = []
            return

        from tqdm import tqdm

        from sonusai.mixture import read_ir
        from sonusai.utils import p_tqdm_map

        progress = tqdm(total=len(self.ir_files), desc='Read impulse response data', disable=not show_progress)
        self._ir_data = p_tqdm_map(read_ir, self.ir_files, progress=progress)
        progress.close()

    def total_samples(self, mixids: GeneralizedIDs = '*') -> int:
        return sum([self.mixture_samples(mixid) for mixid in self.mixids_to_list(mixids)])

    def total_transform_frames(self, mixids: GeneralizedIDs = '*') -> int:
        return self.total_samples(mixids) // self.ft_config.R

    def total_feature_frames(self, mixids: GeneralizedIDs = '*') -> int:
        return self.total_samples(mixids) // self.feature_step_samples

    def mixture_samples(self, mixid: int) -> int:
        return self.mixtures[mixid].samples

    def mixture_transform_frames(self, mixid: int) -> int:
        return self.mixture_samples(mixid) // self.ft_config.R

    def mixture_feature_frames(self, mixid: int) -> int:
        return self.mixture_samples(mixid) // self.feature_step_samples

    def mixids_to_list(self, ids: GeneralizedIDs = None) -> List[int]:
        """Resolve generalized mixture IDs to a list of integers

        :param ids: Generalized mixture IDs
        :return: List of mixture ID integers
        """
        return generic_ids_to_list(len(self.mixtures), ids)

    def mixture_metadata(self, mixid: int) -> str:
        """Create a string of metadata for a mixture ID

        :param mixid: Mixture ID
        :return: String of metadata
        """
        mrecord = self.mixtures[mixid]
        metadata = ''
        for ti in range(len(mrecord.target_file_index)):
            tfi = mrecord.target_file_index[ti]
            tai = mrecord.target_augmentation_index[ti]
            metadata += f'target {ti} name: {self.targets[tfi].name}\n'
            metadata += f'target {ti} augmentation: {self.target_augmentations[tai].to_dict()}\n'
            if self.target_augmentations[tai].ir is None:
                ir_name = None
            else:
                ir_name = self.ir_files[self.target_augmentations[tai].ir]
            metadata += f'target {ti} ir: {ir_name}\n'
            metadata += f'target {ti} target_gain: {mrecord.target_gain[ti]}\n'
            truth_settings = self.targets[tfi].truth_settings
            for tsi in range(len(truth_settings)):
                metadata += f'target {ti} truth index {tsi}: {truth_settings[tsi].index}\n'
                metadata += f'target {ti} truth function {tsi}: {truth_settings[tsi].function}\n'
                metadata += f'target {ti} truth config {tsi}: {truth_settings[tsi].config}\n'
        nfi = mrecord.noise_file_index
        nai = mrecord.noise_augmentation_index
        metadata += f'noise name: {self.noises[nfi].name}\n'
        metadata += f'noise augmentation: {self.noise_augmentations[nai].to_dict()}\n'
        if self.noise_augmentations[nai].ir is None:
            ir_name = None
        else:
            ir_name = self.ir_files[self.noise_augmentations[nai].ir]
        metadata += f'noise ir: {ir_name}\n'
        metadata += f'snr: {mrecord.snr}\n'
        metadata += f'random_snr: {mrecord.random_snr}\n'
        metadata += f'samples: {mrecord.samples}\n'
        metadata += f'target_snr_gain: {mrecord.target_snr_gain}\n'
        metadata += f'noise_snr_gain: {mrecord.noise_snr_gain}\n'

        return metadata

    def write_mixture_metadata(self, mixid: int) -> None:
        """Write mixture metadata to a text file

        :param mixid: Mixture ID
        """
        from os.path import splitext

        with open(file=splitext(self.mixture_filename(mixid))[0] + '.txt', mode='w') as f:
            f.write(self.mixture_metadata(mixid))

    def mixture_filename(self, mixid: int) -> Location:
        """Get the HDF5 file name for the given mixture ID

        :param mixid: Mixture ID
        :return: File name
        """
        from os.path import join

        return join(self.location, self.mixtures[mixid].name) if self.location is not None else None

    def check_audio_files_exist(self) -> None:
        """Walk through all the noise and target audio files in a mixture database ensuring that they exist
        """
        from os.path import exists

        from sonusai import SonusAIError
        from sonusai.mixture import tokenized_expandvars

        for file_index in range(len(self.noises)):
            file_name, _ = tokenized_expandvars(self.noises[file_index].name)
            if not exists(file_name):
                raise SonusAIError(f'Could not find {file_name}')

        for file_index in range(len(self.targets)):
            file_name, _ = tokenized_expandvars(self.targets[file_index].name)
            if not exists(file_name):
                raise SonusAIError(f'Could not find {file_name}')

    def read_mixture_data(self, mixid: int, items: Union[List[str], str]) -> Any:
        """Read mixture data from a mixture HDF5 file

        :param mixid: Mixture ID
        :param items: String(s) of dataset(s) to retrieve
        :return: Data (or tuple of data)
        """
        from os.path import exists

        import h5py
        import numpy as np

        from sonusai import SonusAIError

        def _get_dataset(file: h5py.File, name: str) -> Any:
            if name in file:
                return np.array(file[name])
            return None

        if not isinstance(items, list):
            items = [items]

        name = self.mixture_filename(mixid)
        if exists(name):
            try:
                with h5py.File(name, 'r') as f:
                    result = ([_get_dataset(f, item) for item in items])
            except Exception as e:
                raise SonusAIError(f'Error reading {name}: {e}')
        else:
            result = ([None for item in items])

        if len(items) == 1:
            result = result[0]

        return result

    def write_mixture_data(self, mixid: int, items: Union[List[Tuple[str, Any]], Tuple[str, Any]]) -> None:
        """Write mixture data to a mixture HDF5 file

        :param mixid: Mixture ID
        :param items: Tuple(s) of (name, data)
        """
        import h5py

        if not isinstance(items, list):
            items = [items]

        with h5py.File(self.mixture_filename(mixid), 'a') as f:
            for item in items:
                if item[0] in f:
                    del f[item[0]]
                f.create_dataset(name=item[0], data=item[1])

    def mixture_targets(self,
                        mixid: int,
                        force: bool = False) -> AudiosT:
        """Get the list of augmented target audio data (one per target in the mixup) for the given mixid

        :param mixid: Mixture ID
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: List of augmented target audio data (one per target in the mixup)
        """
        from sonusai.mixture import apply_augmentation
        from sonusai.mixture import apply_gain
        from sonusai.mixture import pad_audio_to_length
        from sonusai.mixture import read_audio

        if not force:
            targets = self.read_mixture_data(mixid, 'targets')
            if targets is not None:
                return list(targets)

        mrecord = self.mixtures[mixid]
        targets = []
        for idx in range(len(mrecord.target_file_index)):
            if self.raw_target_audios is not None:
                target = self.raw_target_audios[mrecord.target_file_index[idx]]
            else:
                target = read_audio(self.targets[mrecord.target_file_index[idx]].name)

            target = apply_augmentation(audio=target,
                                        augmentation=self.target_augmentations[mrecord.target_augmentation_index[idx]],
                                        length_common_denominator=self.feature_step_samples)

            target = apply_gain(audio=target, gain=mrecord.target_snr_gain)
            target = pad_audio_to_length(audio=target, length=mrecord.samples)
            targets.append(target)

        return targets

    def mixture_targets_f(self,
                          mixid: int,
                          targets: AudiosT = None,
                          force: bool = False) -> AudiosF:
        """Get the list of augmented target transform data (one per target in the mixup) for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: List of augmented target transform data (one per target in the mixup)
        """
        if targets is None:
            targets = self.mixture_targets(mixid=mixid, force=force)

        return [self.forward_transform(target) for target in targets]

    def mixture_target(self,
                       mixid: int,
                       targets: AudiosT = None,
                       force: bool = False) -> AudioT:
        """Get the augmented target audio data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Augmented target audio data
        """
        from sonusai.mixture import apply_ir

        if not force:
            target = self.read_mixture_data(mixid, 'target')
            if target is not None:
                return target

        if targets is None:
            targets = self.mixture_targets(mixid=mixid, force=force)

        # Apply impulse responses to targets
        targets_ir = []
        for idx, target in enumerate(targets):
            ir_idx = self.target_augmentations[self.mixtures[mixid].target_augmentation_index[idx]].ir
            if ir_idx is not None:
                targets_ir.append(apply_ir(audio=target, ir=self.ir_data[ir_idx]))
            else:
                targets_ir.append(target)
        targets = targets_ir

        return sum(targets)

    def mixture_target_f(self,
                         mixid: int,
                         targets: AudiosT = None,
                         target: AudioT = None,
                         force: bool = False) -> AudioF:
        """Get the augmented target transform data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Augmented target transform data
        """
        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets, force=force)

        return self.forward_transform(target)

    def mixture_noise(self,
                      mixid: int,
                      force: bool = False) -> AudioT:
        """Get the augmented noise audio data for the given mixid

        :param mixid: Mixture ID
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Augmented noise audio data
        """
        from sonusai.mixture import apply_augmentation
        from sonusai.mixture import apply_gain
        from sonusai.mixture import get_next_noise
        from sonusai.mixture import read_audio

        if not force:
            noise = self.read_mixture_data(mixid, 'noise')
            if noise is not None:
                return noise

        mrecord = self.mixtures[mixid]
        if self.augmented_noise_audios is not None:
            noise = self.augmented_noise_audios[mrecord.noise_file_index][mrecord.noise_augmentation_index]
        else:
            noise = apply_augmentation(audio=read_audio(self.noises[mrecord.noise_file_index].name),
                                       augmentation=self.noise_augmentations[mrecord.noise_augmentation_index])

        noise = get_next_noise(audio=noise, offset=mrecord.noise_offset, length=mrecord.samples)
        noise = apply_gain(audio=noise, gain=mrecord.noise_snr_gain)

        return noise

    def mixture_noise_f(self,
                        mixid: int,
                        noise: AudioT = None,
                        force: bool = False) -> AudioF:
        """Get the augmented noise transform for the given mixid

        :param mixid: Mixture ID
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Augmented noise transform data
        """
        if noise is None:
            noise = self.mixture_noise(mixid=mixid, force=force)

        return self.forward_transform(noise)

    def mixture_mixture(self,
                        mixid: int,
                        targets: AudiosT = None,
                        target: AudioT = None,
                        noise: AudioT = None,
                        force: bool = False) -> AudioT:
        """Get the mixture audio data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Mixture audio data
        """
        if not force:
            mixture = self.read_mixture_data(mixid, 'mixture')
            if mixture is not None:
                return mixture

        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        mixture = target + noise

        return mixture

    def mixture_mixture_f(self, mixid: int,
                          targets: AudiosT = None,
                          target: AudioT = None,
                          noise: AudioT = None,
                          mixture: AudioT = None,
                          force: bool = False) -> AudioF:
        """Get the mixture transform for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param mixture: Mixture audio data for the given mixid
        :return: Mixture transform data
        """
        if mixture is None:
            mixture = self.mixture_mixture(mixid=mixid, targets=targets, target=target, noise=noise, force=force)

        return self.forward_transform(mixture)

    def mixture_truth_t(self,
                        mixid: int,
                        targets: AudiosT = None,
                        noise: AudioT = None,
                        force: bool = False) -> Truth:
        """Get the truth_t data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: truth_t data
        """
        import numpy as np

        from sonusai import SonusAIError
        from sonusai.mixture import TruthFunctionConfig
        from sonusai.mixture import truth_function

        if not force:
            truth_t = self.read_mixture_data(mixid, 'truth_t')
            if truth_t is not None:
                return truth_t

        if targets is None:
            targets = self.mixture_targets(mixid=mixid)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        mrecord = self.mixtures[mixid]
        if len(targets) != len(mrecord.target_file_index):
            raise SonusAIError('Number of target audio entries does not match number of targets')

        if not all(len(target_audio) == len(noise) for target_audio in targets):
            raise SonusAIError('Lengths of target audio do not match length of noise audio')

        truth_t = np.zeros((self.mixture_samples(mixid), self.num_classes), dtype=np.float32)
        for idx in range(len(targets)):
            for truth_setting in self.targets[mrecord.target_file_index[idx]].truth_settings:
                config = TruthFunctionConfig(
                    feature=self.feature,
                    index=truth_setting.index,
                    function=truth_setting.function,
                    config=truth_setting.config,
                    num_classes=self.num_classes,
                    mutex=self.truth_mutex,
                    target_gain=mrecord.target_gain[idx] * mrecord.target_snr_gain
                )
                truth_t += truth_function(target_audio=targets[idx], noise_audio=noise, config=config)

        return truth_t

    def mixture_segsnr_t(self,
                         mixid: int,
                         targets: AudiosT = None,
                         target: AudioT = None,
                         noise: AudioT = None,
                         force: bool = False) -> Segsnr:
        """Get the segsnr_t data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: segsnr_t data
        """
        import numpy as np

        from pyaaware import ForwardTransform

        if not force:
            segsnr_t = self.read_mixture_data(mixid, 'segsnr_t')
            if segsnr_t is not None:
                return segsnr_t

        if target is None:
            target = self.mixture_target(mixid=mixid, targets=targets)

        if noise is None:
            noise = self.mixture_noise(mixid=mixid)

        fft = ForwardTransform(N=self.ft_config.N, R=self.ft_config.R, ttype=self.ft_config.ttype)

        segsnr_t = np.empty(self.mixture_samples(mixid), dtype=np.float32)

        frame = 0
        for offset in range(0, self.mixture_samples(mixid), self.ft_config.R):
            indices = slice(offset, offset + self.ft_config.R)

            target_energy = fft.energy_t(target[indices])
            noise_energy = fft.energy_t(noise[indices])

            if noise_energy == 0:
                snr = np.float32(np.inf)
            else:
                snr = np.float32(target_energy / noise_energy)

            segsnr_t[indices] = snr

        return segsnr_t

    def mixture_segsnr(self,
                       mixid: int,
                       segsnr_t: Segsnr = None,
                       targets: AudiosT = None,
                       target: AudioT = None,
                       noise: AudioT = None,
                       force: bool = False) -> Segsnr:
        """Get the segsnr data for the given mixid

        :param mixid: Mixture ID
        :param segsnr_t: segsnr_t data for the given mixid
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: segsnr data
        """
        if not force:
            segsnr = self.read_mixture_data(mixid, 'segsnr')
            if segsnr is not None:
                return segsnr

            segsnr_t = self.read_mixture_data(mixid, 'segsnr_t')
            if segsnr_t is not None:
                return segsnr_t[0::self.ft_config.R]

        if segsnr_t is None:
            segsnr_t = self.mixture_segsnr_t(mixid=mixid, targets=targets, target=target, noise=noise)

        segsnr = segsnr_t[0::self.ft_config.R]

        return segsnr

    def mixture_ft(self,
                   mixid: int,
                   targets: AudiosT = None,
                   target: AudioT = None,
                   noise: AudioT = None,
                   mixture_f: AudioF = None,
                   mixture: AudioT = None,
                   truth_t: Truth = None,
                   force: bool = False) -> (Feature, Truth):
        """Get the feature and truth_f data for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
        :param target: Augmented target audio data for the given mixid
        :param noise: Augmented noise audio data for the given mixid
        :param mixture_f: Mixture transform data for the given mixid
        :param mixture: Mixture audio data for the given mixid
        :param truth_t: truth_t for the given mixid
        :param force: Force computing data from original sources regardless of whether or not cached data exists
        :return: Tuple of (feature, truth_f) data
        """
        import numpy as np

        from sonusai.mixture import truth_reduction

        if not force:
            feature, truth_f = self.read_mixture_data(mixid, ['feature', 'truth_f'])
            if feature is not None and truth_f is not None:
                return feature, truth_f

        if mixture_f is None:
            mixture_f = self.mixture_mixture_f(mixid=mixid,
                                               targets=targets,
                                               target=target,
                                               noise=noise,
                                               mixture=mixture)

            if truth_t is None:
                truth_t = self.mixture_truth_t(mixid=mixid, targets=targets, noise=noise)

            mrecord = self.mixtures[mixid]

            transform_frames = self.mixture_transform_frames(mixid)
            feature_frames = self.mixture_feature_frames(mixid)

            if truth_t is None:
                truth_t = np.zeros((self.mixture_samples(mixid), self.num_classes), dtype=np.float32)

            feature = np.empty((feature_frames, self.fg.stride, self.fg.num_bands), dtype=np.float32)
            truth_f = np.empty((feature_frames, self.num_classes), dtype=np.complex64)

            feature_frame = 0
            for transform_frame in range(transform_frames):
                indices = slice(transform_frame * self.ft_config.R,
                                (transform_frame + 1) * self.ft_config.R)
                self.fg.execute(mixture_f[transform_frame],
                                truth_reduction(truth_t[indices], self.truth_reduction_function))

                if self.fg.eof():
                    feature[feature_frame] = self.fg.feature()
                    truth_f[feature_frame] = self.fg.truth()
                    feature_frame += 1

            if np.isreal(truth_f).all():
                truth_f = np.float32(np.real(truth_f))

            return feature, truth_f

        def mixture_feature(self,
                            mixid: int,
                            targets: AudiosT = None,
                            noise: AudioT = None,
                            mixture: AudioT = None,
                            truth_t: Truth = None,
                            force: bool = False) -> Feature:
            """Get the feature data for the given mixid

            :param mixid: Mixture ID
            :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
            :param noise: Augmented noise audio data for the given mixid
            :param mixture: Mixture audio data for the given mixid
            :param truth_t: truth_t for the given mixid
            :param force: Force computing data from original sources regardless of whether or not cached data exists
            :return: Feature data
            """
            feature, _ = self.mixture_ft(mixid=mixid,
                                         targets=targets,
                                         noise=noise,
                                         mixture=mixture,
                                         truth_t=truth_t,
                                         force=force)
            return feature

        def mixture_truth_f(self,
                            mixid: int,
                            targets: AudiosT = None,
                            noise: AudioT = None,
                            mixture: AudioT = None,
                            truth_t: Truth = None,
                            force: bool = False) -> Truth:
            """Get the truth_f data for the given mixid

            :param mixid: Mixture ID
            :param targets: List of augmented target audio data (one per target in the mixup) for the given mixid
            :param noise: Augmented noise audio data for the given mixid
            :param mixture: Mixture audio data for the given mixid
            :param truth_t: truth_t for the given mixid
            :param force: Force computing data from original sources regardless of whether or not cached data exists
            :return: truth_f data
            """
            _, truth_f = self.mixture_ft(mixid=mixid,
                                         targets=targets,
                                         noise=noise,
                                         mixture=mixture,
                                         truth_t=truth_t,
                                         force=force)
            return truth_f

    def mixture_class_count(self,
                            mixid: int,
                            targets: AudiosT = None,
                            noise: AudioT = None,
                            truth_t: Truth = None) -> ClassCount:
        """Compute the number of samples for which each truth index is active for the given mixid

        :param mixid: Mixture ID
        :param targets: List of augmented target audio (one per target in the mixup) for the given mixid
        :param noise: Augmented noise audio for the given mixid
        :param truth_t: truth_t for the given mixid
        :return: List of class counts
        """
        if truth_t is None:
            truth_t = self.mixture_truth_t(mixid=mixid, targets=targets, noise=noise)

        class_count = [0] * self.num_classes
        num_classes = self.num_classes
        if self.truth_mutex:
            num_classes -= 1
        for cl in range(num_classes):
            class_count[cl] = int(np.sum(truth_t[:, cl] >= self.class_weights_threshold[cl]))

        return class_count


def get_mrecords_from_mixids(mixdb: MixtureDatabase, mixids: GeneralizedIDs = None) -> MRecords:
    """Get a list of MRecords for the given mixture IDs

    :param mixdb: Mixture database
    :param mixids: Mixture IDs
    :return: MRecords
    """
    from copy import deepcopy

    return [deepcopy(mixdb.mixtures[i]) for i in mixdb.mixids_to_list(mixids)]


def _augment_noise_audio(audio: AudioT) -> AudiosT:
    from sonusai.mixture import apply_augmentation
    from sonusai.mixture import apply_ir

    results = []
    for augmentation in MP_GLOBAL.noise_augmentations:
        result = apply_augmentation(audio, augmentation)
        if augmentation.ir is not None:
            result = apply_ir(result, MP_GLOBAL.ir_data[augmentation.ir])
        results.append(result)

    return results


def generic_ids_to_list(num_ids: int, ids: GeneralizedIDs = None) -> List[int]:
    """Resolve generalized IDs to a list of integers

    :param num_ids: Total number of indices
    :param ids: Generalized  IDs
    :return: List of ID integers
    """
    from sonusai import SonusAIError

    all_ids = list(range(num_ids))

    if ids is None:
        return all_ids

    if isinstance(ids, str):
        if ids == '*':
            return all_ids

        try:
            result = eval(f'{all_ids}[{ids}]')
            if not isinstance(result, list):
                result = [result]
            return result
        except NameError:
            raise SonusAIError(f'Empty ids {ids}')

    if isinstance(ids, range):
        result = list(ids)
    elif isinstance(ids, int):
        result = [ids]
    else:
        result = ids

    if not all(isinstance(x, int) and 0 <= x < num_ids for x in result):
        raise SonusAIError(f'Invalid entries in ids of {ids}')

    if not result:
        raise SonusAIError(f'Empty ids {ids}')

    return result



import numpy as np


def int16_to_float(x: np.ndarray) -> np.ndarray:
    """ Convert int16 array to floating point with range +/- 1
    """
    return np.float32(x) / 32768


def float_to_int16(x: np.ndarray) -> np.ndarray:
    """ Convert float point array with range +/- 1 to int16
    """
    return np.int16(x * 32768)



"""sonusai evaluate

usage: evaluate [-hv] [-i MIXID] (-f FEATURE) (-p PREDICT) [-t PTHR] LOC

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -p PREDICT, --predict PREDICT   A directory containing prediction data.
    -t PTHR, --thr PTHR             Optional prediction decision threshold(s). [default: 0].

Evaluate calculates performance metrics of neural-network models from model prediction data and genft data.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.
    PREDICT     A directory containing SonusAI predict HDF5 files. Contains:
                    dataset:    predict (either [frames, num_classes] or [frames, timesteps, num_classes])
    PTHR        Scalar or array of thresholds. Default 0 will select values:
                    argmax()    if mixdb indicates single-label mode (truth_mutex = true)
                    0.5         if mixdb indicates multi-label mode  (truth_mutex = false)
                If PTHR = -1, optimal thresholds are calculated using precision_recall_curve() which
                optimizes F1 score.
"""
from typing import Union

import numpy as np

from sonusai import logger
from sonusai.mixture import Feature
from sonusai.mixture import Location
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import Predict
from sonusai.mixture import Segsnr
from sonusai.mixture import Truth


def evaluate(mixdb: MixtureDatabase,
             truth: Truth,
             predict: Predict = None,
             segsnr: Segsnr = None,
             output_dir: Location = None,
             predict_thr: Union[float, np.ndarray] = 0.0,
             feature: Feature = None,
             verbose: bool = False) -> None:
    from os.path import join

    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.metrics import calculate_optimal_thresholds
    from sonusai.metrics import class_summary
    from sonusai.metrics import snr_summary
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.queries import get_mixids_from_snr
    from sonusai.utils import get_num_classes_from_predict
    from sonusai.utils import human_readable_size
    from sonusai.utils import reshape_outputs
    from sonusai.utils import seconds_to_hms

    update_console_handler(verbose)
    initial_log_messages('evaluate')

    if truth.shape[-1] != predict.shape[-1]:
        logger.exception(f'Number of classes in truth and predict are not equal. Exiting.')
        raise SystemExit(1)

    # truth, predict can be either [frames, num_classes] or [frames, timesteps, num_classes]
    # in binary case dim may not exist, detect this and set num_classes == 1
    timesteps = -1
    predict, truth = reshape_outputs(predict=predict, truth=truth, timesteps=timesteps)
    num_classes = get_num_classes_from_predict(predict=predict, timesteps=timesteps)

    fdiff = truth.shape[0] - predict.shape[0]
    if fdiff > 0:
        # truth = truth[0:-fdiff,:]
        predict = np.concatenate((predict, np.zeros((fdiff, num_classes))))
        logger.info(f'Truth has more feature-frames than predict, padding predict with zeros to match.')

    if fdiff < 0:
        predict = predict[0:fdiff, :]
        logger.info(f'Predict has more feature-frames than truth, trimming predict to match.')

    # Check segsnr, input is always in transform frames
    compute_segsnr = False
    if len(segsnr) > 0:
        segsnr_feature_frames = segsnr.shape[0] / (mixdb.feature_step_samples / mixdb.fg.ftransform_R)
        if segsnr_feature_frames == truth.shape[0]:
            compute_segsnr = True
        else:
            logger.warning('segsnr length does not match truth, ignoring.')

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb.truth_mutex:
        if num_classes > 1:
            if np.ndim(predict_thr) == 0 and predict_thr == 0:
                # multi-label pred_thr scalar 0 force to 0.5 default
                predict_thr = 0.5

            if np.ndim(predict_thr) == 1:
                if len(predict_thr) == 1:
                    if predict_thr[0] == 0:
                        # multi-label pred_thr array scalar 0 force to 0.5 default
                        predict_thr = 0.5
                    else:
                        # multi-label pred_thr array set to scalar = array[0]
                        predict_thr = predict_thr[0]

    else:
        # single-label mode, force argmax mode
        predict_thr = 0

    if predict_thr == -1:
        thrpr, thrroc, _, _ = calculate_optimal_thresholds(truth, predict, timesteps)
        predict_thr = np.atleast_1d(thrpr)
        predict_thr = np.maximum(predict_thr, 0.001)  # enforce lower limit
        predict_thr = np.minimum(predict_thr, 0.999)  # enforce upper limit
        predict_thr = predict_thr.round(2)

    # Summarize the mixture data
    num_mixtures = len(mixdb.mixtures)
    total_samples = sum([mixture.samples for mixture in mixdb.mixtures])
    duration = total_samples / SAMPLE_RATE

    logger.info('')
    logger.info(f'Mixtures: {num_mixtures}')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'truth:    {human_readable_size(truth.nbytes, 1)}')
    logger.info(f'predict:  {human_readable_size(predict.nbytes, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(segsnr.nbytes, 1)}')
    if feature:
        logger.info(f'feature:  {human_readable_size(feature.nbytes, 1)}')

    logger.info(f'Classes: {num_classes}')
    if mixdb.truth_mutex:
        logger.info(f'Mode:  Single-label / truth_mutex / softmax')
    else:
        logger.info(f'Mode:  Multi-label / Binary')

    mxid_snro = get_mixids_from_snr(mixdb=mixdb)
    snrlist = list(mxid_snro.keys())
    snrlist.sort(reverse=True)
    logger.info(f'Ordered SNRs: {snrlist}\n')
    if type(predict_thr) is np.ndarray:
        logger.info(f'Prediction Threshold(s): {predict_thr.transpose()}\n')
    else:
        logger.info(f'Prediction Threshold(s): {predict_thr}\n')

    # Top-level report over all mixtures
    macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb=mixdb,
                                                      mixid=':',
                                                      truth_f=truth,
                                                      predict=predict,
                                                      segsnr=segsnr if compute_segsnr else None,
                                                      predict_thr=predict_thr)

    if num_classes > 1:
        logger.info(f'Metrics micro-avg per SNR over all {num_mixtures} mixtures:')
    else:
        logger.info(f'Metrics per SNR over all {num_mixtures} mixtures:')
    logger.info(microdf.round(3).to_string())
    logger.info('')
    if output_dir:
        microdf.round(3).to_csv(join(output_dir, 'snr.csv'))

    if mixdb.truth_mutex:
        macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb=mixdb,
                                                          mixid=':',
                                                          truth_f=truth[:, 0:-1],
                                                          predict=predict[:, 0:-1],
                                                          segsnr=segsnr if compute_segsnr else None,
                                                          predict_thr=predict_thr)

        logger.info(f'Metrics micro-avg without "Other" class per SNR over all {num_mixtures} mixtures:')
        logger.info(microdf.round(3).to_string())
        logger.info('')
        if output_dir:
            microdf.round(3).to_csv(join(output_dir, 'snrwo.csv'))

    for snri in snrlist:
        mxids = mxid_snro[snri]
        classdf = class_summary(mixdb, mxids, truth, predict, predict_thr)
        logger.info(f'Metrics per class for SNR {snri} over {len(mxids)} mixtures:')
        logger.info(classdf.round(3).to_string())
        logger.info('')
        if output_dir:
            classdf.round(3).to_csv(join(output_dir, f'class_snr{snri}.csv'))


def main():
    from datetime import datetime
    from os import mkdir
    from os.path import join

    import h5py
    from docopt import docopt

    import sonusai
    from sonusai import SonusAIError
    from sonusai import create_file_handler
    from sonusai.utils import read_predict_data
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    feature_name = args['--feature']
    predict_name = args['--predict']
    predict_threshold = float(args['--thr'])
    location = args['LOC']

    mixdb = MixtureDatabase(config=location, show_progress=True)

    # create output directory
    output_dir = f'evaluate-{datetime.now():%Y%m%d}'
    try:
        mkdir(output_dir)
    except OSError as _:
        output_dir = f'evaluate-{datetime.now():%Y%m%d-%H%M%S}'
        try:
            mkdir(output_dir)
        except OSError as error:
            raise SonusAIError(f'Could not create directory, {output_dir}: {error}')

    create_file_handler(join(output_dir, 'evaluate.log'))

    with h5py.File(feature_name, 'r') as f:
        truth_f = np.array(f['truth_f'])
        segsnr = np.array(f['segsnr'])

    predict = read_predict_data(predict_name)

    evaluate(mixdb=mixdb,
             truth=truth_f,
             segsnr=segsnr,
             output_dir=output_dir,
             predict=predict,
             predict_thr=predict_threshold,
             verbose=verbose)

    logger.info(f'Wrote results to {output_dir}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



# SonusAI mixture utilities
from sonusai.mixture.audio import calculate_audio_from_transform
from sonusai.mixture.audio import calculate_transform_from_audio
from sonusai.mixture.audio import get_next_noise
from sonusai.mixture.audio import read_audio
from sonusai.mixture.audio import read_ir
from sonusai.mixture.augmentation import apply_augmentation
from sonusai.mixture.augmentation import apply_gain
from sonusai.mixture.augmentation import apply_ir
from sonusai.mixture.augmentation import estimate_audio_length
from sonusai.mixture.augmentation import evaluate_random_rule
from sonusai.mixture.augmentation import get_augmentation_indices_for_mixup
from sonusai.mixture.augmentation import get_augmentations
from sonusai.mixture.augmentation import get_mixups
from sonusai.mixture.augmentation import get_pad_length
from sonusai.mixture.augmentation import pad_audio_to_length
from sonusai.mixture.balance import get_class_balancing_augmentation
from sonusai.mixture.class_count import get_class_count_from_mixids
from sonusai.mixture.config import generate_mixture_filename
from sonusai.mixture.config import get_class_weights_threshold
from sonusai.mixture.config import get_default_config
from sonusai.mixture.config import get_ir_files
from sonusai.mixture.config import get_max_class
from sonusai.mixture.config import get_noise_files
from sonusai.mixture.config import get_target_files
from sonusai.mixture.config import load_config
from sonusai.mixture.config import raw_load_config
from sonusai.mixture.config import tokenized_expandvars
from sonusai.mixture.config import update_config_from_file
from sonusai.mixture.config import update_config_from_hierarchy
from sonusai.mixture.config import update_truth_settings
from sonusai.mixture.constants import BIT_DEPTH
from sonusai.mixture.constants import CHANNEL_COUNT
from sonusai.mixture.constants import DEFAULT_CONFIG
from sonusai.mixture.constants import DEFAULT_NOISE
from sonusai.mixture.constants import ENCODING
from sonusai.mixture.constants import FLOAT_BYTES
from sonusai.mixture.constants import RAND_PATTERN
from sonusai.mixture.constants import REQUIRED_CONFIGS
from sonusai.mixture.constants import SAMPLE_BYTES
from sonusai.mixture.constants import SAMPLE_RATE
from sonusai.mixture.constants import VALID_AUGMENTATIONS
from sonusai.mixture.constants import VALID_CONFIGS
from sonusai.mixture.constants import VALID_TRUTH_SETTINGS
from sonusai.mixture.feature import get_feature_from_audio
from sonusai.mixture.initialize import initialize_mixture
from sonusai.mixture.initialize import initialize_target
from sonusai.mixture.log_duration_and_sizes import log_duration_and_sizes
from sonusai.mixture.mapped_snr_f import calculate_mapped_snr_f
from sonusai.mixture.mapped_snr_f import calculate_snr_f_statistics
from sonusai.mixture.mixdb import MRecord
from sonusai.mixture.mixdb import MRecords
from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.mixdb import MixtureDatabaseConfig
from sonusai.mixture.mixdb import generic_ids_to_list
from sonusai.mixture.mixdb import get_mrecords_from_mixids
from sonusai.mixture.target_class_balancing import balance_targets
from sonusai.mixture.targets import get_augmented_target_indices_by_class
from sonusai.mixture.targets import get_augmented_target_indices_for_mixup
from sonusai.mixture.targets import get_augmented_targets
from sonusai.mixture.targets import get_target_indices_for_truth_index
from sonusai.mixture.targets import get_truth_indices_for_target
from sonusai.mixture.truth import get_truth_indices_for_mixid
from sonusai.mixture.truth import truth_function
from sonusai.mixture.truth import truth_reduction
from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosF
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import Augmentation
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import AugmentedTarget
from sonusai.mixture.types import AugmentedTargets
from sonusai.mixture.types import Feature
from sonusai.mixture.types import GenFTData
from sonusai.mixture.types import GenMixData
from sonusai.mixture.types import GeneralizedIDs
from sonusai.mixture.types import ImpulseResponseData
from sonusai.mixture.types import ImpulseResponseFiles
from sonusai.mixture.types import ListAudiosT
from sonusai.mixture.types import Location
from sonusai.mixture.types import Predict
from sonusai.mixture.types import Segsnr
from sonusai.mixture.types import TargetFile
from sonusai.mixture.types import TargetFiles
from sonusai.mixture.types import Truth
from sonusai.mixture.types import TruthFunctionConfig
from sonusai.mixture.types import TruthSettings



import os
import re

import sonusai

REQUIRED_CONFIGS = [
    'class_balancing',
    'class_balancing_augmentation',
    'class_labels',
    'class_weights_threshold',
    'exhaustive_noise',
    'feature',
    'impulse_responses',
    'noise_augmentations',
    'noises',
    'num_classes',
    'random_snrs',
    'seed',
    'snrs',
    'target_augmentations',
    'targets',
    'truth_mode',
    'truth_reduction_function',
    'truth_settings',
]
VALID_CONFIGS = REQUIRED_CONFIGS + ['output']
VALID_TRUTH_SETTINGS = ['function', 'config', 'index']
VALID_AUGMENTATIONS = ['normalize', 'gain', 'pitch', 'tempo', 'eq1', 'eq2', 'eq3', 'lpf', 'ir', 'count', 'mixup']
RAND_PATTERN = re.compile(r'rand\(([-+]?(\d+(\.\d*)?|\.\d+)),\s*([-+]?(\d+(\.\d*)?|\.\d+))\)')
SAMPLE_RATE = 16000
BIT_DEPTH = 32
ENCODING = 'floating-point'
CHANNEL_COUNT = 1
SAMPLE_BYTES = BIT_DEPTH // 8
FLOAT_BYTES = 4

DEFAULT_NOISE = os.path.join(sonusai.BASEDIR, 'data', 'whitenoise.wav')
DEFAULT_CONFIG = os.path.join(sonusai.BASEDIR, 'data', 'genmixdb.yml')



from __future__ import annotations

from decimal import Decimal
from typing import Any
from typing import Union


class EngineeringNumber:
    """Easy manipulation of numbers which use engineering notation"""

    _suffix_lookup = {
        'Y': 'e24',
        'Z': 'e21',
        'E': 'e18',
        'P': 'e15',
        'T': 'e12',
        'G': 'e9',
        'M': 'e6',
        'k': 'e3',
        '': 'e0',
        'm': 'e-3',
        'u': 'e-6',
        'n': 'e-9',
        'p': 'e-12',
        'f': 'e-15',
        'a': 'e-18',
        'z': 'e-21',
        'y': 'e-24',
    }

    _exponent_lookup_scaled = {
        '-12': 'Y',
        '-15': 'Z',
        '-18': 'E',
        '-21': 'P',
        '-24': 'T',
        '-27': 'G',
        '-30': 'M',
        '-33': 'k',
        '-36': '',
        '-39': 'm',
        '-42': 'u',
        '-45': 'n',
        '-48': 'p',
        '-51': 'f',
        '-54': 'a',
        '-57': 'z',
        '-60': 'y',
    }

    def __init__(self,
                 value: Union[str, float, int, EngineeringNumber],
                 precision: int = 2,
                 significant: int = 0):
        """
        :param value: string, integer, or float representing the numeric value of the number
        :param precision: the precision past the decimal
        :param significant: the number of significant digits
        if given, significant takes precedence over precision
        """
        self.precision = precision
        self.significant = significant

        if isinstance(value, str):
            suffix_keys = [key for key in self._suffix_lookup.keys() if key != '']

            for suffix in suffix_keys:
                if suffix in value:
                    value = value[:-1] + self._suffix_lookup[suffix]
                    break

            self.number = Decimal(value)

        elif isinstance(value, int) or isinstance(value, float) or isinstance(value, EngineeringNumber):
            self.number = Decimal(str(value))

        else:
            raise TypeError('value has unsupported type')

    def __repr__(self):
        """Returns the string representation"""
        # The Decimal class only really converts numbers that are very small into engineering notation.
        # So we will simply make all numbers small numbers and take advantage of the Decimal class.
        number_str = self.number * Decimal('10e-37')
        number_str = number_str.to_eng_string().lower()

        base, exponent = number_str.split('e')

        if self.significant > 0:
            abs_base = abs(Decimal(base))
            num_digits = 1
            num_digits += 1 if abs_base >= 10 else 0
            num_digits += 1 if abs_base >= 100 else 0
            num_digits = self.significant - num_digits
        else:
            num_digits = self.precision

        base = str(round(Decimal(base), num_digits))

        if 'e' in base.lower():
            base = str(int(Decimal(base)))

        # Remove trailing decimal
        if '.' in base:
            base = base.rstrip('.')

        return base + self._exponent_lookup_scaled[exponent]

    def __str__(self) -> str:
        return self.__repr__()

    def __int__(self) -> int:
        return int(self.number)

    def __float__(self):
        return float(self.number)

    @staticmethod
    def _to_decimal(other: Union[str, float, int, EngineeringNumber]) -> Decimal:
        if not isinstance(other, EngineeringNumber):
            other = EngineeringNumber(other)
        return other.number

    def __add__(self, other: Union[str, float, int, EngineeringNumber]) -> EngineeringNumber:
        return EngineeringNumber(str(self.number + self._to_decimal(other)))

    def __radd__(self, other: Union[str, float, int, EngineeringNumber]) -> EngineeringNumber:
        return self.__add__(other)

    def __sub__(self, other: Union[str, float, int, EngineeringNumber]) -> EngineeringNumber:
        return EngineeringNumber(str(self.number - self._to_decimal(other)))

    def __rsub__(self, other: Union[str, float, int, EngineeringNumber]) -> EngineeringNumber:
        return EngineeringNumber(str(self._to_decimal(other) - self.number))

    def __mul__(self, other: Union[str, float, int, EngineeringNumber]) -> EngineeringNumber:
        return EngineeringNumber(str(self.number * self._to_decimal(other)))

    def __rmul__(self, other: Union[str, float, int, EngineeringNumber]) -> EngineeringNumber:
        return self.__mul__(other)

    def __truediv__(self, other: Union[str, float, int, EngineeringNumber]) -> EngineeringNumber:
        return EngineeringNumber(str(self.number / self._to_decimal(other)))

    def __rtruediv__(self, other: Union[str, float, int, EngineeringNumber]) -> EngineeringNumber:
        return EngineeringNumber(str(self._to_decimal(other) / self.number))

    def __lt__(self, other: Union[str, float, int, EngineeringNumber]) -> bool:
        return self.number < self._to_decimal(other)

    def __gt__(self, other: Union[str, float, int, EngineeringNumber]) -> bool:
        return self.number > self._to_decimal(other)

    def __le__(self, other: Union[str, float, int, EngineeringNumber]) -> bool:
        return self.number <= self._to_decimal(other)

    def __ge__(self, other: Union[str, float, int, EngineeringNumber]) -> bool:
        return self.number >= self._to_decimal(other)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, (str, float, int, EngineeringNumber)):
            return NotImplemented
        return self.number == self._to_decimal(other)



# SonusAI query utilities
from sonusai.queries.queries import get_mixids_from_noise
from sonusai.queries.queries import get_mixids_from_noise_augmentation
from sonusai.queries.queries import get_mixids_from_snr
from sonusai.queries.queries import get_mixids_from_target
from sonusai.queries.queries import get_mixids_from_target_augmentation
from sonusai.queries.queries import get_mixids_from_truth_function
from sonusai.queries.queries import get_mixids_from_truth_index



"""sonusai predict

usage: predict [-hvr] [-i MIXID] (-m MODEL) INPUT

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to generate if input is a mixture database. [default: *].
    -m MODEL, --model MODEL     Trained ONNX model file.
    -r, --reset                 Reset model between each file.

Run prediction on a trained ONNX model using SonusAI genft or WAV data.

Inputs:
    MODEL       A SonusAI trained ONNX model file.
    INPUT       The input data must be one of the following:
                * WAV
                  Using the given model, generate feature data and run prediction. A model file must be
                  provided. The MIXID is ignored.

                * directory
                  Using the given SonusAI mixture database directory, generate feature and truth data if not found.
                  Run prediction. The MIXID is required.

Outputs the following to opredict-<TIMESTAMP> directory:
    <id>.h5
        dataset:    predict
    onnx_predict.log

"""
from sonusai import logger
from sonusai.mixture import Feature
from sonusai.mixture import Predict
from sonusai.utils import SonusAIMetaData


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixids = args['--mixid']
    model_name = args['--model']
    reset = args['--reset']
    input_name = args['INPUT']

    from os import makedirs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    import h5py
    import onnxruntime as rt
    import numpy as np

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import get_feature_from_audio
    from sonusai.mixture import read_audio
    from sonusai.utils import create_ts_name
    from sonusai.utils import get_frames_per_batch
    from sonusai.utils import get_sonusai_metadata

    output_dir = create_ts_name('opredict')
    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    create_file_handler(join(output_dir, 'onnx_predict.log'))
    update_console_handler(verbose)
    initial_log_messages('onnx_predict')

    model = rt.InferenceSession(model_name, providers=['CPUExecutionProvider'])
    model_metadata = get_sonusai_metadata(model)

    batch_size = model_metadata.input_shape[0]
    if model_metadata.timestep:
        timesteps = model_metadata.input_shape[1]
    else:
        timesteps = 0
    num_classes = model_metadata.output_shape[-1]

    frames_per_batch = get_frames_per_batch(batch_size, timesteps)

    logger.info('')
    logger.info(f'feature       {model_metadata.feature}')
    logger.info(f'num_classes   {num_classes}')
    logger.info(f'batch_size    {batch_size}')
    logger.info(f'timesteps     {timesteps}')
    logger.info(f'flatten       {model_metadata.flattened}')
    logger.info(f'add1ch        {model_metadata.channel}')
    logger.info(f'truth_mutex   {model_metadata.mutex}')
    logger.info(f'input_shape   {model_metadata.input_shape}')
    logger.info(f'output_shape  {model_metadata.output_shape}')
    logger.info('')

    if splitext(input_name)[1] == '.wav':
        # Convert WAV to feature data
        logger.info('')
        logger.info(f'Run prediction on {input_name}')
        audio = read_audio(input_name)
        feature = get_feature_from_audio(audio=audio, feature=model_metadata.feature)

        predict = pad_and_predict(feature=feature,
                                  model_name=model_name,
                                  model_metadata=model_metadata,
                                  frames_per_batch=frames_per_batch,
                                  batch_size=batch_size,
                                  timesteps=timesteps,
                                  reset=reset)

        output_name = splitext(input_name)[0] + '.h5'
        with h5py.File(output_name, 'a') as f:
            if 'feature' in f:
                del f['feature']
            f.create_dataset(name='feature', data=feature)

            if 'predict' in f:
                del f['predict']
            f.create_dataset(name='predict', data=predict)

        logger.info(f'Saved results to {output_name}')
        return

    if not isdir(input_name):
        logger.exception(f'Do not know how to process input from {input_name}')
        raise SystemExit(1)

    mixdb = MixtureDatabase(config=input_name, show_progress=True)

    if mixdb.feature != model_metadata.feature:
        logger.exception(f'Feature in mixture database does not match feature in model')
        raise SystemExit(1)

    mixids = mixdb.mixids_to_list(mixids)
    if reset:
        # reset mode cycles through each file one at a time
        for mixid in mixids:
            feature, _ = mixdb.mixture_ft(mixid)

            predict = pad_and_predict(feature=feature,
                                      model_name=model_name,
                                      model_metadata=model_metadata,
                                      frames_per_batch=frames_per_batch,
                                      batch_size=batch_size,
                                      timesteps=timesteps,
                                      reset=reset)

            output_name = join(output_dir, mixdb.mixtures[mixid].name)
            with h5py.File(output_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset(name='predict', data=predict)
    else:
        feature = []
        file_indices = []
        total_frames = 0
        for mixid in mixids:
            current_feature, _ = mixdb.mixture_ft(mixid)
            current_frames = current_feature.shape[0]
            feature.append(current_feature)
            file_indices.append(slice(total_frames, total_frames + current_frames))
            total_frames += current_frames
        feature = np.vstack([feature[i] for i in range(len(feature))])

        predict = pad_and_predict(feature=feature,
                                  model_name=model_name,
                                  model_metadata=model_metadata,
                                  frames_per_batch=frames_per_batch,
                                  batch_size=batch_size,
                                  timesteps=timesteps,
                                  reset=reset)

        # Write data to separate files
        for idx, mixid in enumerate(mixids):
            output_name = join(output_dir, mixdb.mixtures[mixid].name)
            with h5py.File(output_name, 'a') as f:
                if 'predict' in f:
                    del f['predict']
                f.create_dataset('predict', data=predict[file_indices[idx]])

    logger.info(f'Saved results to {output_dir}')


def pad_and_predict(feature: Feature,
                    model_name: str,
                    model_metadata: SonusAIMetaData,
                    frames_per_batch: int,
                    batch_size: int,
                    timesteps: int,
                    reset: bool) -> Predict:
    import onnxruntime as rt
    import numpy as np

    from sonusai.utils import reshape_inputs
    from sonusai.utils import reshape_outputs

    frames = feature.shape[0]
    padding = frames_per_batch - frames % frames_per_batch
    feature = np.pad(array=feature, pad_width=((0, padding), (0, 0), (0, 0)))
    feature, _ = reshape_inputs(feature=feature,
                                batch_size=batch_size,
                                timesteps=timesteps,
                                flatten=model_metadata.flattened,
                                add1ch=model_metadata.channel)
    sequences = feature.shape[0] // model_metadata.input_shape[0]
    feature = np.reshape(feature, [sequences, *model_metadata.input_shape])

    model = rt.InferenceSession(model_name, providers=['CPUExecutionProvider'])
    output_names = [n.name for n in model.get_outputs()]
    input_names = [n.name for n in model.get_inputs()]

    predict = []
    for sequence in range(sequences):
        predict.append(model.run(output_names, {input_names[0]: feature[sequence]}))
        if reset:
            model = rt.InferenceSession(model_name, providers=['CPUExecutionProvider'])

    predict = np.vstack(predict)
    predict, _ = reshape_outputs(predict=predict, timesteps=timesteps)
    predict = predict[:frames, :]

    return predict


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



import warnings
from typing import List

import numpy as np

from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import MixtureDatabase

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from keras.utils import Sequence


class KerasFromMixtureDatabase(Sequence):
    """Generates data for Keras from a SonusAI mixture database
    """
    from dataclasses import dataclass

    @dataclass(frozen=True)
    class BatchParams:
        mixids: List[int]
        offset: int
        extra: int
        padding: int

    def __init__(self,
                 mixdb: MixtureDatabase,
                 mixids: GeneralizedIDs,
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False):
        """Initialization
        """
        self.mixdb = mixdb
        self.mixids = self.mixdb.mixids_to_list(mixids)
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle
        self.stride = self.mixdb.fg.stride
        self.num_bands = self.mixdb.fg.num_bands
        self.num_classes = self.mixdb.num_classes
        self.mixture_frame_segments = None
        self.batch_frame_segments = None
        self.total_batches = None

        self._initialize_mixtures()

    def __len__(self) -> int:
        """Denotes the number of batches per epoch
        """
        return self.total_batches

    def __getitem__(self, batch_index: int) -> (np.ndarray, np.ndarray):
        """Get one batch of data
        """
        from sonusai.utils import reshape_inputs

        batch_params = self.batch_params[batch_index]

        result = [self.mixdb.mixture_ft(mixid) for mixid in batch_params.mixids]
        feature = np.vstack([result[i][0] for i in range(len(result))])
        truth = np.vstack([result[i][1] for i in range(len(result))])

        pad_shape = list(feature.shape)
        pad_shape[0] = batch_params.padding
        feature = np.vstack([feature, np.zeros(pad_shape)])

        pad_shape = list(truth.shape)
        pad_shape[0] = batch_params.padding
        truth = np.vstack([truth, np.zeros(pad_shape)])

        if batch_params.extra > 0:
            feature = feature[batch_params.offset:-batch_params.extra]
            truth = truth[batch_params.offset:-batch_params.extra]
        else:
            feature = feature[batch_params.offset:]
            truth = truth[batch_params.offset:]

        feature, truth = reshape_inputs(feature=feature,
                                        truth=truth,
                                        batch_size=self.batch_size,
                                        timesteps=self.timesteps,
                                        flatten=self.flatten,
                                        add1ch=self.add1ch)

        return feature, truth

    def on_epoch_end(self) -> None:
        """Modification of dataset between epochs
        """
        import random

        if self.shuffle:
            random.shuffle(self.mixids)
            self._initialize_mixtures()

    def _initialize_mixtures(self) -> None:
        from sonusai.utils import get_frames_per_batch

        frames_per_batch = get_frames_per_batch(self.batch_size, self.timesteps)
        # Always extend the number of batches to use all available data
        # The last batch may need padding
        self.total_batches = int(np.ceil(self.mixdb.total_feature_frames(self.mixids) / frames_per_batch))

        # Compute mixid, offset, and extra for dataset
        # offsets and extras are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence. And the extra must be maintained in order to preserve the correct data length.
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        self.batch_params = []
        self.file_indices = []
        total_frames = 0
        for idx, mixid in enumerate(self.mixids):
            current_frames = self.mixdb.mixture_samples(mixid) // self.mixdb.feature_step_samples
            self.file_indices.append(slice(total_frames, total_frames + current_frames))
            total_frames += current_frames
            cumulative_frames += current_frames
            while cumulative_frames >= frames_per_batch:
                extra = cumulative_frames - frames_per_batch
                mixids = self.mixids[start_mixture_index:idx + 1]
                self.batch_params.append(self.BatchParams(mixids=mixids, offset=offset, extra=extra, padding=0))
                if extra == 0:
                    start_mixture_index = idx + 1
                    offset = 0
                else:
                    start_mixture_index = idx
                    offset = current_frames - extra
                cumulative_frames = extra

        # If needed, add final batch with padding
        needed_frames = self.total_batches * frames_per_batch
        padding = needed_frames - total_frames
        if padding != 0:
            mixids = self.mixids[start_mixture_index:]
            self.batch_params.append(self.BatchParams(mixids=mixids, offset=offset, extra=0, padding=padding))


KerasFromH5 = KerasFromMixtureDatabase



from typing import List


def expand_braces(text: str, seen: set = None) -> List[str]:
    import itertools
    import re

    if seen is None:
        seen = set()

    spans = [m.span() for m in re.finditer('\{[^\{\}]*\}', text)][::-1]
    alts = [text[start + 1: stop - 1].split(',') for start, stop in spans]

    if len(spans) == 0:
        if text not in seen:
            yield text
        seen.add(text)
    else:
        for combo in itertools.product(*alts):
            replaced = list(text)
            for (start, stop), replacement in zip(spans, combo):
                replaced[start:stop] = replacement
            yield from expand_braces(''.join(replaced), seen)


def braced_glob(path: str) -> List[str]:
    from glob import glob

    result = []
    for x in expand_braces(path):
        result.extend(glob(x))

    return result



import numpy as np


def calculate_sample_weights(class_weights: np.ndarray, truth: np.ndarray) -> np.ndarray:
    """Calculate sample weights from class weights and a given truth with 2D or 3D shape.

    Supports one-hot encoded multi-class or binary truth/labels
    Note returns sum of weighted truth over classes, thus should also work for multi-label ? TBD

    Inputs:
      class_weights [num_classes, 1]        weights for each class
      truth         [frames, timesteps, num_classes] or [frames, num_classes]

    Returns:
      sample_weights [frames, timesteps, 1] or [frames, 1]
    """
    ts = truth.shape
    cs = class_weights.shape

    if ts[-1] == 1 and cs[0] == 2:
        # Binary truth needs 2nd "none" truth dimension
        truth = np.concatenate((truth, 1 - truth), axis=1)

    # broadcast [num_classes, 1] over [frames, num_classes] or [frames, timesteps, num_classes]
    return np.sum(class_weights * truth, axis=-1)



import logging
from os.path import dirname

from pkg_resources import get_distribution

__version__ = get_distribution('sonusai').version
BASEDIR = dirname(__file__)

# create logger
logger = logging.getLogger('sonusai')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(message)s')
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


class SonusAIError(Exception):
    def __init__(self, value):
        logger.error(value)


# create file handler
def create_file_handler(filename: str):
    fh = logging.FileHandler(filename=filename, mode='w')
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


# update console handler
def update_console_handler(verbose: bool):
    if not verbose:
        logger.removeHandler(console_handler)
        console_handler.setLevel(logging.INFO)
        logger.addHandler(console_handler)


# write initial log message
def initial_log_messages(name: str):
    from datetime import datetime
    from getpass import getuser
    from os import getcwd
    from socket import gethostname
    from sys import argv

    logger.info(f'SonusAI {__version__}')
    logger.info(f'{name}')
    logger.info('')
    logger.debug(f'Host:      {gethostname()}')
    logger.debug(f'User:      {getuser()}')
    logger.debug(f'Directory: {getcwd()}')
    logger.debug(f'Date:      {datetime.now()}')
    logger.debug(f'Command:   {" ".join(argv)}')
    logger.debug('')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sonusai',
 'sonusai.data_generator',
 'sonusai.metrics',
 'sonusai.mixture',
 'sonusai.queries',
 'sonusai.utils']

package_data = \
{'': ['*'], 'sonusai': ['data/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'docopt>=0.6.2,<0.7.0',
 'keras-tuner>=1.1.3,<2.0.0',
 'matplotlib>=3.6.1,<4.0.0',
 'onnxruntime-gpu>=1.12.1,<2.0.0',
 'paho-mqtt>=1.6.1,<2.0.0',
 'pandas>=1.5.1,<2.0.0',
 'pesq>=0.0.4,<0.0.5',
 'pyaaware>=1.4.7,<2.0.0',
 'scikit-learn>=1.2.0,<2.0.0',
 'setuptools>=65.6.0,<66.0.0',
 'sh>=1.14.3,<2.0.0',
 'sox>=1.4.1,<2.0.0',
 'tensorflow-addons>=0.18.0,<0.19.0',
 'tensorflow>=2.9.0,<2.10.0',
 'tf2onnx>=1.12.1,<2.0.0',
 'tqdm>=4.64.1,<5.0.0']

entry_points = \
{'console_scripts': ['aawscd_probwrite = sonusai.aawscd_probwrite:main',
                     'sonusai = sonusai.main:main']}

setup_kwargs = {
    'name': 'sonusai',
    'version': '0.10.11',
    'description': 'Framework for building deep neural network models for sound, speech, and voice AI',
    'long_description': "Sonus AI: Framework for simplified creation of deep NN models for sound, speech, and voice AI\n\nSonus AI includes functions for pre-processing training and validation data and\ncreating performance metrics reports for key types of Keras models:\n- recurrent, convolutional, or a combination (i.e. RCNNs)\n- binary, multiclass single-label, multiclass multi-label, and regresssion\n- training with data augmentations:  noise mixing, pitch and time stretch, etc.\n\nSonus AI python functions are used by:\n - Aaware Inc. sonusai executable:  Easily create train/validation data, run prediction, evaluate model performance\n - Keras model scripts:             User python scripts for keras model creation, training, and prediction. These can use sonusai-specific data but also some general useful utilities for trainining rnn-based models like CRNN's, DSCRNN's, etc. in Keras\n",
    'author': 'Chris Eddington',
    'author_email': 'chris@aaware.com',
    'maintainer': 'Chris Eddington',
    'maintainer_email': 'chris@aaware.com',
    'url': 'https://aaware.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from sonusai.mixture import SAMPLE_RATE
from sonusai.mixture.types import AudioT


def write_wav(name: str, audio: AudioT, sample_rate: int = SAMPLE_RATE) -> None:
    """ Write a simple, uncompressed WAV file.

    To write multiple channels, use a 2D array of shape [samples, channels].
    The bits per sample and PCM/float are determined by the data type.

    """
    from scipy.io.wavfile import write

    write(name, sample_rate, audio)



# SonusAI metrics utilities for model training and validation
from sonusai.metrics.calculate_class_weights import calculate_class_weights_from_mixdb
from sonusai.metrics.calculate_class_weights import calculate_class_weights_from_truth
from sonusai.metrics.calculate_optimal_thresholds import calculate_optimal_thresholds
from sonusai.metrics.calculate_sample_weights import calculate_sample_weights
from sonusai.metrics.class_summary import class_summary
from sonusai.metrics.confusion_matrix_summary import confusion_matrix_summary
from sonusai.metrics.one_hot import one_hot
from sonusai.metrics.snr_summary import snr_summary



from typing import List

import numpy as np

from sonusai.mixture.mixdb import MixtureDatabase
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import Truth
from sonusai.mixture.types import TruthFunctionConfig


def _strictly_decreasing(list_to_check: list) -> bool:
    return all(x > y for x, y in zip(list_to_check, list_to_check[1:]))


def truth_function(target_audio: AudioT, noise_audio: AudioT, config: TruthFunctionConfig) -> Truth:
    import h5py
    from pyaaware import FeatureGenerator
    from pyaaware import ForwardTransform
    from pyaaware import SED

    from sonusai import SonusAIError
    from sonusai.mixture import calculate_mapped_snr_f

    fg = FeatureGenerator(feature_mode=config.feature, num_classes=config.num_classes, truth_mutex=config.mutex)
    frame_size = fg.ftransform_R
    truth = np.zeros((len(target_audio), config.num_classes), dtype=np.float32)
    offsets = range(0, len(target_audio), frame_size)
    zero_based_indices = [x - 1 for x in config.index]
    target_fft = ForwardTransform(N=fg.ftransform_N, R=fg.ftransform_R, ttype=fg.ftransform_ttype)
    noise_fft = ForwardTransform(N=fg.ftransform_N, R=fg.ftransform_R, ttype=fg.ftransform_ttype)

    if config.function == 'sed':
        if len(target_audio) % frame_size != 0:
            raise SonusAIError(f'Number of samples in audio is not a multiple of {frame_size}')

        if config.config is None:
            raise SonusAIError('Truth function SED missing config')

        parameters = ['thresholds']
        for parameter in parameters:
            if parameter not in config.config:
                raise SonusAIError(f'Truth function SED config missing required parameter: {parameter}')

        thresholds = config.config['thresholds']
        if not _strictly_decreasing(thresholds):
            raise SonusAIError(f'Truth function SED thresholds are not strictly decreasing: {thresholds}')

        if config.target_gain == 0:
            return truth

        # SED wants 1-based indices
        sed = SED(thresholds=thresholds,
                  index=config.index,
                  frame_size=frame_size,
                  num_classes=config.num_classes,
                  mutex=config.mutex)

        target_audio = target_audio / config.target_gain
        for offset in offsets:
            indices = slice(offset, offset + frame_size)
            new_truth = sed.execute(target_fft.energy_t(target_audio[indices]))
            truth[indices] = np.reshape(new_truth, (1, len(new_truth)))

        return truth

    if config.function == 'file':
        if config.config is None:
            raise SonusAIError('Truth function file missing config')

        parameters = ['file']
        for parameter in parameters:
            if parameter not in config.config:
                raise SonusAIError(f'Truth function file config missing required parameter: {parameter}')

        with h5py.File(config.config['file'], 'r') as f:
            truth_in = np.array(f['truth_t'])

        if truth_in.ndim != 2:
            raise SonusAIError('Truth file data is not 2 dimensions')

        if truth_in.shape[0] != len(target_audio):
            raise SonusAIError('Truth file does not contain the right amount of samples')

        if config.target_gain == 0:
            return truth

        if len(zero_based_indices) > 1:
            if len(zero_based_indices) != truth_in.shape[1]:
                raise SonusAIError('Truth file does not contain the right amount of classes')

            truth[:, zero_based_indices] = truth_in
        else:
            index = zero_based_indices[0]
            if index + truth_in.shape[1] > config.num_classes:
                raise SonusAIError('Truth file contains too many classes')

            truth[:, index:index + truth_in.shape[1]] = truth_in

        return truth

    if config.function == 'energy_t':
        if config.target_gain == 0:
            return truth

        for offset in offsets:
            target_energy = target_fft.energy_t(target_audio[offset:offset + frame_size])
            truth[offset:offset + frame_size, zero_based_indices] = np.float32(target_energy)

        return truth

    if config.function == 'target_f':
        if config.num_classes != 2 * target_fft.bins:
            raise SonusAIError(f'Invalid num_classes for target_f truth: {config.num_classes}')

        truth = np.zeros((len(target_audio), config.num_classes), dtype=np.float32)
        if config.target_gain == 0:
            return truth

        for offset in offsets:
            target_f = np.complex64(target_fft.execute(target_audio[offset:offset + frame_size]))

            indices = slice(offset, offset + frame_size)
            for index in zero_based_indices:
                start = index
                stop = start + target_fft.bins
                truth[indices, start:stop] = np.real(target_f)

                start = stop
                stop = start + target_fft.bins
                truth[indices, start:stop] = np.imag(target_f)

        return truth

    if config.function == 'target_mixture_f':
        if config.num_classes != 2 * target_fft.bins + 2 * noise_fft.bins:
            raise SonusAIError(f'Invalid num_classes for target_mixture_f truth: {config.num_classes}')

        truth = np.zeros((len(target_audio), config.num_classes), dtype=np.float32)
        if config.target_gain == 0:
            return truth

        for offset in offsets:
            target_f = np.complex64(target_fft.execute(target_audio[offset:offset + frame_size]))
            noise_f = np.complex64(noise_fft.execute(noise_audio[offset:offset + frame_size]))
            mixture_f = target_f + noise_f

            indices = slice(offset, offset + frame_size)
            for index in zero_based_indices:
                start = index
                stop = start + target_fft.bins
                truth[indices, start:stop] = np.real(target_f)

                start = stop
                stop = start + target_fft.bins
                truth[indices, start:stop] = np.imag(target_f)

                start = stop
                stop = start + noise_fft.bins
                truth[indices, start:stop] = np.real(mixture_f)

                start = stop
                stop = start + noise_fft.bins
                truth[indices, start:stop] = np.imag(mixture_f)

        return truth

    if config.function == 'crm':
        if config.num_classes != target_fft.bins:
            raise SonusAIError(f'Invalid num_classes for crm truth: {config.num_classes}')

        if target_fft.bins != noise_fft.bins:
            raise SonusAIError('Transform size mismatch for crm truth')

        truth = np.zeros((len(target_audio), config.num_classes), dtype=np.float32)

        if config.target_gain == 0:
            return truth

        for offset in offsets:
            target_f = np.complex64(target_fft.execute(target_audio[offset:offset + frame_size]))
            noise_f = np.complex64(noise_fft.execute(noise_audio[offset:offset + frame_size]))
            mixture_f = target_f + noise_f

            crm = np.empty(target_f.shape, dtype=np.complex64)
            with np.nditer(target_f, flags=['multi_index'], op_flags=['readwrite']) as it:
                for _ in it:
                    num = target_f[it.multi_index]
                    den = mixture_f[it.multi_index]
                    if num == 0:
                        crm[it.multi_index] = 0
                    elif den == 0:
                        crm[it.multi_index] = complex(np.inf, np.inf)
                    else:
                        crm[it.multi_index] = num / den

            indices = slice(offset, offset + frame_size)
            for index in zero_based_indices:
                truth[indices, index:index + target_fft.bins] = np.real(crm)
                truth[indices, (index + target_fft.bins):(index + 2 * target_fft.bins)] = np.imag(crm)

        return truth

    if config.function in ['energy_f', 'snr_f', 'mapped_snr_f']:
        if config.target_gain == 0:
            return truth

        snr_db_mean = None
        snr_db_std = None
        if config.function == 'mapped_snr_f':
            if config.config is None:
                raise SonusAIError('Truth function mapped SNR missing config')

            parameters = ['snr_db_mean', 'snr_db_std']
            for parameter in parameters:
                if parameter not in config.config:
                    raise SonusAIError(f'Truth function mapped SNR config missing required parameter: {parameter}')

            snr_db_mean = config.config['snr_db_mean']
            if len(snr_db_mean) != target_fft.bins:
                raise SonusAIError(f'Truth function mapped SNR snr_db_mean does not have {target_fft.bins} elements')

            snr_db_std = config.config['snr_db_std']
            if len(snr_db_std) != target_fft.bins:
                raise SonusAIError(f'Truth function mapped SNR snr_db_std does not have {target_fft.bins} elements')

        for index in zero_based_indices:
            if index + target_fft.bins > config.num_classes:
                raise SonusAIError('Truth index exceeds the number of classes')

        for offset in offsets:
            target_energy = np.float32(
                target_fft.energy_f(target_audio[offset:offset + frame_size]))
            if config.function in ['snr_f', 'mapped_snr_f']:
                noise_energy = np.float32(noise_fft.energy_f(noise_audio[offset:offset + frame_size]))
            else:
                noise_energy = np.float32(1)

            indices = slice(offset, offset + frame_size)
            for index in zero_based_indices:
                old_err = np.seterr(divide='ignore', invalid='ignore')
                tmp = target_energy / noise_energy
                np.seterr(**old_err)

                tmp = np.nan_to_num(tmp, nan=-np.inf, posinf=np.inf, neginf=-np.inf)

                if config.function == 'mapped_snr_f':
                    tmp = calculate_mapped_snr_f(tmp, snr_db_mean, snr_db_std)

                truth[indices, index:index + target_fft.bins] = tmp

        return truth

    if config.function == 'phoneme':
        # Read in .txt transcript and run a Python function to generate text grid data
        # (indicating which phonemes are active)
        # Then generate truth based on this data and put in the correct classes based on config.index
        raise SonusAIError('Truth function phoneme is not supported yet')

    raise SonusAIError(f'Unsupported truth function: {config.function}')


def get_truth_indices_for_mixid(mixdb: MixtureDatabase, mixid: int) -> List[int]:
    """Get a list of truth indices for a given mixid."""

    from sonusai.mixture.targets import get_truth_indices_for_target

    indices = []
    for target_file_index in mixdb.mixtures[mixid].target_file_index:
        indices.append(*get_truth_indices_for_target(mixdb.targets[target_file_index]))

    return sorted(list(set(indices)))


def truth_reduction(x: Truth, func: str) -> Truth:
    from sonusai import SonusAIError

    if func == 'max':
        return np.max(x, axis=0)

    if func == 'mean':
        return np.mean(x, axis=0)

    raise SonusAIError(f'Invalid truth reduction function: {func}')



from typing import Dict
from typing import List
from typing import Union

from sonusai.mixture.types import AudioT
from sonusai.mixture.types import Augmentation
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import ImpulseResponseData


def get_augmentations(rules: Union[List[Dict], Dict], num_ir: int = 0) -> Augmentations:
    """Generate augmentations from list of input rules

    :param rules: Dictionary of augmentation config rule[s]
    :param num_ir: Number of impulse responses in config
    :return: List of augmentations
    """
    from sonusai.mixture import Augmentation
    from sonusai.utils import dataclass_from_dict

    processed_rules = []
    if not isinstance(rules, list):
        rules = [rules]

    for rule in rules:
        rule = _parse_ir(rule, num_ir)
        expand_rules(expanded_rules=processed_rules, rule=rule)

    processed_rules = randomize_rules(rules=processed_rules, num_ir=num_ir)

    return [dataclass_from_dict(Augmentation, processed_rule) for processed_rule in processed_rules]


def expand_rules(expanded_rules: List[Dict], rule: dict) -> None:
    """Expand rules

    :param expanded_rules: Working list of expanded rules
    :param rule: Rule to process
    """
    from copy import deepcopy
    from numbers import Number

    from sonusai import SonusAIError
    from sonusai.mixture import VALID_AUGMENTATIONS

    for key, value in list(rule.items()):
        if value is None:
            del rule[key]

    # replace old 'eq' rule with new 'eq1' rule to allow both for backward compatibility
    rule = {'eq1' if key == 'eq' else key: value for key, value in rule.items()}

    for key in rule:
        if key not in VALID_AUGMENTATIONS:
            nice_list = '\n'.join([f'  {item}' for item in VALID_AUGMENTATIONS])
            raise SonusAIError(f'Invalid augmentation: {key}.\nValid augmentations are:\n{nice_list}')

        if key in ['eq1', 'eq2', 'eq3']:
            # EQ must be a list of length 3 or a list of length 3 lists
            valid = True
            multiple = False
            if isinstance(rule[key], list):
                if any(isinstance(el, list) for el in rule[key]):
                    multiple = True
                    for value in rule[key]:
                        if not isinstance(value, list) or len(value) != 3:
                            valid = False
                else:
                    if len(rule[key]) != 3:
                        valid = False
            else:
                valid = False

            if not valid:
                raise SonusAIError(f'Invalid augmentation value for {key}: {rule[key]}')

            if multiple:
                for value in rule[key]:
                    expanded_rule = deepcopy(rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_rules(expanded_rules, expanded_rule)
                return

        elif key in ['count', 'mixup']:
            pass

        else:
            if isinstance(rule[key], list):
                for value in rule[key]:
                    if isinstance(value, list):
                        raise SonusAIError(f'Invalid augmentation value for {key}: {rule[key]}')
                    expanded_rule = deepcopy(rule)
                    expanded_rule[key] = deepcopy(value)
                    expand_rules(expanded_rules, expanded_rule)
                return
            elif not isinstance(rule[key], Number):
                if not rule[key].startswith('rand'):
                    raise SonusAIError(f'Invalid augmentation value for {key}: {rule[key]}')

    expanded_rules.append(rule)


def randomize_rules(rules: List[Dict], num_ir: int = 0) -> List[Dict]:
    """Randomize rules

    :param rules: List of rules to randomize
    :param num_ir: Number of impulse responses in config
    :return: List of randomized rules
    """
    out_rules = []
    for in_rule in rules:
        if rule_has_rand(in_rule):
            count = 1
            if 'count' in in_rule and in_rule['count'] is not None:
                count = in_rule['count']
                del in_rule['count']
            for i in range(count):
                out_rules.append(generate_random_rule(in_rule, num_ir))
        else:
            out_rules.append(in_rule)
    return out_rules


def generate_random_rule(rule: dict, num_ir: int = 0) -> dict:
    """Generate a new rule from a rule that contains 'rand' directives

    :param rule: Rule to randomize
    :param num_ir: Number of impulse responses in config
    :return: Randomized rule
    """
    from copy import deepcopy
    from random import randint

    out_rule = deepcopy(rule)
    for key in out_rule:
        if key == 'ir' and out_rule[key] == 'rand':
            # IR is special case
            if num_ir == 0:
                out_rule[key] = None
            else:
                out_rule[key] = randint(0, num_ir - 1)
        else:
            out_rule[key] = evaluate_random_rule(str(out_rule[key]))

        # convert EQ values from strings to numbers
        if key in ['eq1', 'eq2', 'eq3']:
            for n in range(3):
                if isinstance(out_rule[key][n], str):
                    out_rule[key][n] = eval(out_rule[key][n])

    return out_rule


def rule_has_rand(rule: dict) -> bool:
    """Determine if any keys in the given rule contain 'rand'

    :param rule: Rule to check
    :return: True if rule contains 'rand'
    """
    for key in rule:
        if 'rand' in str(rule[key]):
            return True

    return False


def apply_augmentation(audio: AudioT, augmentation: Augmentation, length_common_denominator: int = 1) -> AudioT:
    """Use sox to apply augmentations to audio data

    :param audio: Audio to augment
    :param augmentation: Augmentation rule
    :param length_common_denominator: Pad resulting audio to be a multiple of this
    :return: Augmented audio
    """
    import sox

    from sonusai import SonusAIError
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import ENCODING
    from sonusai.mixture import SAMPLE_RATE

    try:
        # Apply augmentations
        tfm = sox.Transformer()
        tfm.set_input_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT, encoding=ENCODING)
        tfm.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT, encoding=ENCODING)

        # TODO
        #  Always normalize and remove normalize from list of available augmentations
        #  Normalize to globally set level (should this be a global config parameter,
        #  or hard-coded into the script?)
        if augmentation.normalize is not None:
            tfm.norm(db_level=augmentation.normalize)

        if augmentation.gain is not None:
            tfm.gain(gain_db=augmentation.gain, normalize=False)

        if augmentation.pitch is not None:
            tfm.pitch(n_semitones=augmentation.pitch / 100)

        if augmentation.tempo is not None:
            factor = augmentation.tempo
            if abs(factor - 1.0) <= 0.1:
                tfm.stretch(factor=factor)
            else:
                tfm.tempo(factor=factor, audio_type='s')

        if augmentation.eq1 is not None:
            tfm.equalizer(frequency=augmentation.eq1[0], width_q=augmentation.eq1[1],
                          gain_db=augmentation.eq1[2])

        if augmentation.eq2 is not None:
            tfm.equalizer(frequency=augmentation.eq2[0], width_q=augmentation.eq2[1],
                          gain_db=augmentation.eq2[2])

        if augmentation.eq3 is not None:
            tfm.equalizer(frequency=augmentation.eq3[0], width_q=augmentation.eq3[1],
                          gain_db=augmentation.eq3[2])

        if augmentation.lpf is not None:
            tfm.lowpass(frequency=augmentation.lpf)

        # Create output data
        audio_out = tfm.build_array(input_array=audio, sample_rate_in=SAMPLE_RATE)

        # make sure length is multiple of length_common_denominator
        return _pad_audio(audio=audio_out, length_common_denominator=length_common_denominator)
    except Exception as e:
        raise SonusAIError(f'Error applying {augmentation}: {e}')


def apply_ir(audio: AudioT, ir: ImpulseResponseData) -> AudioT:
    """Use sox to apply impulse response to audio data

    :param audio: Audio to augment
    :param ir: Impulse response data
    :return: Augmented audio
    """
    if ir is None:
        return audio

    import numpy as np
    import sox

    from sonusai.mixture import SAMPLE_RATE

    # Get current maximum level in dB
    max_db = 20 * np.log10(max(abs(audio)))

    # Convert audio to IR sample rate and normalize to -20 dBFS to avoid clipping when applying IR
    tfm = sox.Transformer()
    tfm.set_output_format(rate=ir.sample_rate)
    tfm.norm(db_level=-20)
    audio_out = tfm.build_array(input_array=audio, sample_rate_in=SAMPLE_RATE)

    # Pad audio to align with original and give enough room for IR tail
    pad = int(np.ceil(len(ir.filter) / 2))
    audio_out = np.pad(array=audio_out, pad_width=(pad, pad))

    # Apply IR and convert back to global sample rate
    tfm = sox.Transformer()
    tfm.set_output_format(rate=SAMPLE_RATE)
    tfm.fir(coefficients=ir.filter)
    audio_out = tfm.build_array(input_array=audio_out, sample_rate_in=ir.sample_rate)

    # Reset level to previous max value
    tfm = sox.Transformer()
    tfm.norm(db_level=max_db)
    audio_out = tfm.build_array(input_array=audio_out, sample_rate_in=SAMPLE_RATE)

    # Trim to IR offset to align with input
    audio_out = audio_out[ir.offset:]

    return audio_out[:len(audio)]


def estimate_audio_length(audio: AudioT, augmentation: Augmentation, length_common_denominator: int = 1) -> int:
    """Estimate the length of audio after augmentation

    :param audio: Audio to estimate
    :param augmentation: Augmentation rule
    :param length_common_denominator: Pad resulting audio to be a multiple of this
    :return: Estimated length of augmented audio
    """
    length = len(audio)

    if augmentation.tempo is not None:
        length = int(length // augmentation.tempo)

    length += get_pad_length(length, length_common_denominator)

    return length


def get_mixups(augmentations: Augmentations) -> List[int]:
    """Get a list of mixup values used

    :param augmentations: List of augmentations
    :return: List of mixup values used
    """
    return sorted(list(set([augmentation.mixup for augmentation in augmentations])))


def get_augmentation_indices_for_mixup(augmentations: Augmentations, mixup: int) -> List[int]:
    """Get a list of augmentation indices for a given mixup value

    :param augmentations: List of augmentations
    :param mixup: Mixup value of interest
    :return: List of augmentation indices
    """
    indices = []
    for idx, augmentation in enumerate(augmentations):
        if mixup == augmentation.mixup:
            indices.append(idx)

    return indices


def _pad_audio(audio: AudioT, length_common_denominator: int = 1) -> AudioT:
    """Pad audio to be a multiple of given value

    :param audio: Audio to pad
    :param length_common_denominator: Pad resulting audio to be a multiple of this
    :return: Padded audio
    """
    import numpy as np

    return np.pad(array=audio, pad_width=(0, get_pad_length(len(audio), length_common_denominator)))


def get_pad_length(length: int, length_common_denominator: int) -> int:
    """Get the number of pad samples needed

    :param length: Length of original
    :param length_common_denominator: Desired length will be a multiple of this
    :return: Number of pad samples required
    """
    mod = int(length % length_common_denominator)
    return length_common_denominator - mod if mod else 0


def pad_audio_to_length(audio: AudioT, length: int) -> AudioT:
    """Pad audio to given length

    :param audio: Audio to pad
    :param length: Length of output
    :return: Padded audio
    """
    import numpy as np

    return np.pad(array=audio, pad_width=(0, length - len(audio)))


def apply_gain(audio: AudioT, gain: float) -> AudioT:
    """Apply gain to audio

    :param audio: Audio to adjust
    :param gain: Amount of gain
    :return: Adjusted audio
    """
    return audio * gain


def evaluate_random_rule(rule: str) -> Union[str, float]:
    """Evaluate 'rand' directive

    :param rule: Rule to evaluate
    :return: Resolved value
    """
    import re
    from random import uniform

    from sonusai.mixture.constants import RAND_PATTERN

    def rand_repl(m):
        return f'{uniform(float(m.group(1)), float(m.group(4))):.2f}'

    return eval(re.sub(RAND_PATTERN, rand_repl, rule))


def _parse_ir(rule: dict, num_ir: int) -> dict:
    from sonusai import SonusAIError
    from sonusai.mixture import generic_ids_to_list

    if 'ir' not in rule:
        return rule

    ir = rule['ir']

    if ir is None:
        return rule

    if isinstance(ir, str):
        if ir == 'rand':
            return rule

        rule['ir'] = generic_ids_to_list(num_ir, ir)
        return rule

    if isinstance(ir, list):
        if not all(item in range(num_ir) for item in ir):
            raise SonusAIError(f'Invalid ir of {ir}')
        return rule

    if isinstance(ir, int):
        if ir not in range(num_ir):
            raise SonusAIError(f'Invalid ir of {ir}')
        return rule

    raise SonusAIError(f'Invalid ir of {ir}')



"""sonusai mkwav

usage: mkwav [-hvtn] [-i MIXID] LOC

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -t, --target                    Write target file.
    -n, --noise                     Write noise file.

The mkwav command creates WAV files from a SonusAI database.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.

Outputs the following to the mixture database directory:
    <id>_mixture.wav:   mixture
    <id>_target.wav:    target (optional)
    <id>_noise.wav:     noise (optional)
    <id>.txt
    mkwav.log

"""
from dataclasses import dataclass

from sonusai import logger
from sonusai.mixture import AudioT
from sonusai.mixture import MixtureDatabase


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class MPGlobal:
    mixdb: MixtureDatabase = None
    write_target: bool = None
    write_noise: bool = None


MP_GLOBAL = MPGlobal()


def mkwav(mixdb: MixtureDatabase, mixid: int) -> (AudioT, AudioT, AudioT):
    from sonusai.genmix import genmix

    data = genmix(mixdb=mixdb, mixids=mixid, force=False)

    return data[0].mixture, sum(data[0].targets), data[0].noise


def _process_mixture(mixid: int) -> None:
    from os.path import exists
    from os.path import join
    from os.path import splitext

    import h5py
    import numpy as np

    from sonusai.utils import float_to_int16
    from sonusai.utils import write_wav

    mixture_filename = join(MP_GLOBAL.mixdb.location, MP_GLOBAL.mixdb.mixtures[mixid].name)
    mixture_basename = splitext(mixture_filename)[0]

    target = None
    noise = None

    need_data = True
    if exists(mixture_filename + '.h5'):
        with h5py.File(mixture_filename, 'r') as f:
            if 'mixture' in f:
                need_data = False
            if MP_GLOBAL.write_target and 'targets' not in f:
                need_data = True
            if MP_GLOBAL.write_noise and 'noise' not in f:
                need_data = True

    if need_data:
        mixture, target, noise = mkwav(mixdb=MP_GLOBAL.mixdb, mixid=mixid)
    else:
        with h5py.File(mixture_filename, 'r') as f:
            mixture = np.array(f['mixture'])
            if MP_GLOBAL.write_target:
                target = sum(np.array(f['targets']))
            if MP_GLOBAL.write_noise:
                noise = np.array(f['noise'])

    write_wav(name=mixture_basename + '_mixture.wav', audio=float_to_int16(mixture))
    if MP_GLOBAL.write_target:
        write_wav(name=mixture_basename + '_target.wav', audio=float_to_int16(target))
    if MP_GLOBAL.write_noise:
        write_wav(name=mixture_basename + '_noise.wav', audio=float_to_int16(noise))

    with open(file=mixture_basename + '.txt', mode='w') as f:
        f.write(MP_GLOBAL.mixdb.mixture_metadata(mixid))


def main():
    import time
    from os.path import join

    from docopt import docopt
    from tqdm import tqdm

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    mixid = args['--mixid']
    write_target = args['--target']
    write_noise = args['--noise']
    location = args['LOC']

    start_time = time.monotonic()

    create_file_handler(join(location, 'mkwav.log'))
    update_console_handler(verbose)
    initial_log_messages('mkwav')

    logger.info(f'Load mixture database from {location}')
    mixdb = MixtureDatabase(config=location, show_progress=True)
    mixid = mixdb.mixids_to_list(mixid)

    total_samples = mixdb.total_samples(mixid)
    duration = total_samples / sonusai.mixture.SAMPLE_RATE

    logger.info('')
    logger.info(f'Found {len(mixid):,} mixtures to process')
    logger.info(f'{total_samples:,} samples')

    mixdb.check_audio_files_exist()

    MP_GLOBAL.mixdb = mixdb
    MP_GLOBAL.write_target = write_target
    MP_GLOBAL.write_noise = write_noise

    progress = tqdm(total=len(mixid), desc='mkwav')
    p_tqdm_map(_process_mixture, mixid, progress=progress)
    progress.close()

    logger.info(f'Wrote {len(mixid)} mixtures to {location}')
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'mixture:  {human_readable_size(total_samples * 2, 1)}')
    if write_target:
        logger.info(f'target:   {human_readable_size(total_samples * 2, 1)}')
    if write_noise:
        logger.info(f'noise:    {human_readable_size(total_samples * 2, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')
    logger.info('')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from typing import Tuple

import numpy as np

from sonusai.mixture.types import Feature
from sonusai.mixture.types import Predict
from sonusai.mixture.types import Truth


def get_input_shape(feature: Feature) -> Tuple[int]:
    return feature.shape[1:]


def reshape_inputs(feature: Feature,
                   batch_size: int,
                   truth: Truth = None,
                   timesteps: int = 0,
                   flatten: bool = False,
                   add1ch: bool = False) -> (Feature, Truth):
    """Check SonusAI feature and truth data and reshape feature of size [frames, strides, bands] into
    one of several options:

    If timesteps > 0: (i.e., for recurrent NNs):
      no-flatten, no-channel:   [sequences, timesteps, strides, bands]      (4-dim)
      flatten, no-channel:      [sequences, timesteps, strides*bands]       (3-dim)
      no-flatten, add-1channel: [sequences, timesteps, strides, bands, 1]   (5-dim)
      flatten, add-1channel:    [sequences, timesteps, strides*bands, 1]    (4-dim)

    If batch_size is None, then do not reshape; just calculate new input shape and return.

    If timesteps == 0, then do not add timesteps dimension.

    The number of samples is trimmed to be a multiple of batch_size (Keras requirement) for
    both feature and truth.
    Channel is added to last/outer dimension for channel_last support in Keras/TF.

    Returns:
      feature       reshaped feature
      truth         reshaped truth
    """
    from sonusai import SonusAIError

    frames, strides, bands = feature.shape
    if truth is not None:
        truth_frames, num_classes = truth.shape
        # Double-check correctness of inputs
        if frames != truth_frames:
            raise SonusAIError('Frames in feature and truth do not match')
    else:
        num_classes = None

    if flatten:
        feature = np.reshape(feature, (frames, strides * bands))

    # Reshape for Keras/TF recurrent models that require timesteps/sequence length dimension
    if timesteps > 0:
        sequences = frames // timesteps

        # Remove frames if remainder exists (not fitting into a multiple of new number of sequences)
        frames_rem = frames % timesteps
        batch_rem = (frames // timesteps) % batch_size
        bf_rem = batch_rem * timesteps
        sequences = sequences - batch_rem
        fr2drop = frames_rem + bf_rem
        if fr2drop:
            if feature.ndim == 2:
                feature = feature[0:-fr2drop, ]  # flattened input
            elif feature.ndim == 3:
                feature = feature[0:-fr2drop, ]  # un-flattened input

            if truth is not None:
                truth = truth[0:-fr2drop, ]

        # Reshape
        if feature.ndim == 2:  # flattened input
            # str=str+'Reshaping 2 dim\n'
            feature = np.reshape(feature, (sequences, timesteps, strides * bands))  # was [frames, bands*timesteps]
            if truth is not None:
                truth = np.reshape(truth, (sequences, timesteps, num_classes))  # was [frames, num_classes]
        elif feature.ndim == 3:  # un-flattened input
            # str=str+'Reshaping 3 dim\n'
            feature = np.reshape(feature, (sequences, timesteps, strides, bands))  # was [frames, bands, timesteps]
            if truth is not None:
                truth = np.reshape(truth, (sequences, timesteps, num_classes))  # was [frames, num_classes]
    else:
        # Drop frames if remainder exists (not fitting into a multiple of new number of sequences)
        fr2drop = feature.shape[0] % batch_size
        if fr2drop > 0:
            feature = feature[0:-fr2drop, ]
            if truth is not None:
                truth = truth[0:-fr2drop, ]

    # Add channel dimension if required for input to model (i.e. for cnn type input)
    if add1ch:
        feature = np.expand_dims(feature, axis=feature.ndim)  # add as last/outermost dim

    return feature, truth


def get_num_classes_from_predict(predict: Predict, timesteps: int = 0) -> int:
    num_dims = predict.ndim
    dims = predict.shape

    if num_dims == 3 or (num_dims == 2 and timesteps > 0):
        # 2D with timesteps - [frames, timesteps]
        if num_dims == 2:
            return 1

        # 3D - [frames, timesteps, num_classes]
        return dims[2]

    # 1D - [frames]
    if num_dims == 1:
        return 1

    # 2D without timesteps - [frames, num_classes]
    return dims[1]


def reshape_outputs(predict: Predict, truth: Truth = None, timesteps: int = 0) -> (Predict, Truth):
    """Reshape model output data.

    truth and predict can be either [frames, num_classes], or [frames, timesteps, num_classes]
    In binary case, num_classes dim may not exist; detect this and set num_classes to 1.
    """
    from sonusai import SonusAIError

    if truth is not None:
        if predict.shape != truth.shape:
            raise SonusAIError('predict and truth shapes do not match')

    ndim = predict.ndim
    shape = predict.shape

    if not (0 < ndim <= 3):
        raise SonusAIError(f'do not know how to reshape data with {ndim} dimensions')

    if ndim == 3 or (ndim == 2 and timesteps > 0):
        if ndim == 2:
            # 2D with timesteps - [frames, timesteps]
            num_classes = 1
        else:
            # 3D - [frames, timesteps, num_classes]
            num_classes = shape[2]

        # reshape to remove timestep dimension
        shape = (shape[0] * shape[1], num_classes)
        predict = np.reshape(predict, shape)
        if truth is not None:
            truth = np.reshape(truth, shape)
    elif ndim == 1:
        # convert to 2D - [frames, 1]
        predict = np.expand_dims(predict, 1)
        if truth is not None:
            truth = np.expand_dims(truth, 1)

    return predict, truth
